@@ -1,12 +1,19 @@
 /*	ruby mysql module
- *	$Id: mysql.c,v 1.23 2000/09/02 17:34:47 tommy Exp $
+ *	$Id: mysql.c,v 1.32 2002/01/07 01:17:28 tommy Exp $
  */
 
 #include "ruby.h"
+#ifdef HAVE_MYSQL_H
+#include <mysql.h>
+#include <errmsg.h>
+#else
 #include <mysql/mysql.h>
 #include <mysql/errmsg.h>
+#endif
 
-#ifndef str2cstr		/* ruby 1.2.x ? */
+#define GC_STORE_RESULT_LIMIT 20
+
+#ifndef Qtrue		/* ruby 1.2.x ? */
 #define	Qtrue		TRUE
 #define	Qfalse		FALSE
 #define	rb_exc_raise	rb_raise
@@ -38,6 +45,8 @@ VALUE cMysqlRes;
 VALUE cMysqlField;
 VALUE eMysql;
 
+static int store_result_count = 0;
+
 struct mysql {
     MYSQL handler;
     char connection;
@@ -60,8 +69,10 @@ static void free_mysql(struct mysql* my)
 
 static void free_mysqlres(struct mysql_res* resp)
 {
-    if (resp->freed == Qfalse)
+    if (resp->freed == Qfalse) {
 	mysql_free_result(resp->res);
+	store_result_count--;
+    }
     free(resp);
 }
 
@@ -80,6 +91,8 @@ static VALUE mysqlres2obj(MYSQL_RES* res)
     resp->res = res;
     resp->freed = Qfalse;
     rb_obj_call_init(obj, 0, NULL);
+    if (++store_result_count > GC_STORE_RESULT_LIMIT)
+	rb_gc();
     return obj;
 }
 
@@ -125,8 +138,7 @@ static VALUE real_connect(int argc, VALUE* argv, VALUE klass)
 {
     VALUE host, user, passwd, db, port, sock, flag;
     char *h, *u, *p, *d, *s;
-    uint pp, f;
-    MYSQL my;
+    unsigned int pp, f;
     struct mysql* myp;
     VALUE obj;
 
@@ -146,18 +158,17 @@ static VALUE real_connect(int argc, VALUE* argv, VALUE klass)
     pp = NILorINT(port);
     s = NILorSTRING(sock);
 
-#if MYSQL_VERSION_ID >= 32200
-    mysql_init(&my);
-    if (mysql_real_connect(&my, h, u, p, d, pp, s, f) == NULL)
-#elif MYSQL_VERSION_ID >= 32115
-    if (mysql_real_connect(&my, h, u, p, pp, s, f) == NULL)
-#else
-    if (mysql_real_connect(&my, h, u, p, pp, s) == NULL)
-#endif
-	mysql_raise(&my);
-
     obj = Data_Make_Struct(klass, struct mysql, 0, free_mysql, myp);
-    myp->handler = my;
+#if MYSQL_VERSION_ID >= 32200
+    mysql_init(&myp->handler);
+    if (mysql_real_connect(&myp->handler, h, u, p, d, pp, s, f) == NULL)
+#elif MYSQL_VERSION_ID >= 32115
+    if (mysql_real_connect(&myp->handler, h, u, p, pp, s, f) == NULL)
+#else
+    if (mysql_real_connect(&myp->handler, h, u, p, pp, s) == NULL)
+#endif
+	mysql_raise(&myp->handler);
+
     myp->connection = Qtrue;
     myp->query_with_result = Qtrue;
     rb_obj_call_init(obj, argc, argv);
@@ -181,6 +192,15 @@ static VALUE client_info(VALUE klass)
     return rb_tainted_str_new2(mysql_get_client_info());
 }
 
+#if MYSQL_VERSION_ID >= 32332
+/*	my_debug(string)	*/
+static VALUE my_debug(VALUE obj, VALUE str)
+{
+    mysql_debug(STR2CSTR(str));
+    return obj;
+}
+#endif
+
 /*-------------------------------
  * Mysql object method
  */
@@ -191,7 +211,7 @@ static VALUE real_connect2(int argc, VALUE* argv, VALUE obj)
 {
     VALUE host, user, passwd, db, port, sock, flag;
     char *h, *u, *p, *d, *s;
-    uint pp, f;
+    unsigned int pp, f;
     MYSQL* m = GetHandler(obj);
     rb_scan_args(argc, argv, "07", &host, &user, &passwd, &db, &port, &sock, &flag);
     d = NILorSTRING(db);
@@ -241,6 +261,19 @@ static VALUE options(int argc, VALUE* argv, VALUE obj)
 }
 #endif
 
+#if MYSQL_VERSION_ID >= 32332
+/*	real_escape_string(string)	*/
+static VALUE real_escape_string(VALUE obj, VALUE str)
+{
+    MYSQL* m = GetHandler(obj);
+    VALUE ret;
+    Check_Type(str, T_STRING);
+    ret = rb_str_new(0, (RSTRING(str)->len)*2+1);
+    RSTRING(ret)->len = mysql_real_escape_string(m, RSTRING(ret)->ptr, RSTRING(str)->ptr, RSTRING(str)->len);
+    return ret;
+}
+#endif
+
 /*	initialize()	*/
 static VALUE initialize(int argc, VALUE* argv, VALUE obj)
 {
@@ -253,6 +286,31 @@ static VALUE affected_rows(VALUE obj)
     return INT2NUM(mysql_affected_rows(GetHandler(obj)));
 }
 
+#if MYSQL_VERSION_ID >= 32303
+/*	change_user(user=nil, passwd=nil, db=nil)	*/
+static VALUE change_user(int argc, VALUE* argv, VALUE obj)
+{
+    VALUE user, passwd, db;
+    char *u, *p, *d;
+    MYSQL* m = GetHandler(obj);
+    rb_scan_args(argc, argv, "03", &user, &passwd, &db);
+    u = NILorSTRING(user);
+    p = NILorSTRING(passwd);
+    d = NILorSTRING(db);
+    if (mysql_change_user(m, u, p, d) != 0)
+	mysql_raise(m);
+    return obj;
+}
+#endif
+
+#if MYSQL_VERSION_ID >= 32321
+/*	character_set_name()	*/
+static VALUE character_set_name(VALUE obj)
+{
+    return rb_tainted_str_new2(mysql_character_set_name(GetHandler(obj)));
+}
+#endif
+
 /*	close()		*/
 static VALUE my_close(VALUE obj)
 {
@@ -264,6 +322,7 @@ static VALUE my_close(VALUE obj)
     return obj;
 }
 
+#if MYSQL_VERSION_ID < 40000
 /*	create_db(db)	*/
 static VALUE create_db(VALUE obj, VALUE db)
 {
@@ -281,6 +340,18 @@ static VALUE drop_db(VALUE obj, VALUE db)
 	mysql_raise(m);
     return obj;
 }
+#endif
+
+#if MYSQL_VERSION_ID >= 32332
+/*	dump_debug_info()	*/
+static VALUE dump_debug_info(VALUE obj)
+{
+    MYSQL* m = GetHandler(obj);
+    if (mysql_dump_debug_info(m) != 0)
+	mysql_raise(m);
+    return obj;
+}
+#endif
 
 /*	errno()		*/
 static VALUE my_errno(VALUE obj)
@@ -692,6 +763,7 @@ static VALUE res_free(VALUE obj)
     check_free(obj);
     mysql_free_result(resp->res);
     resp->freed = Qtrue;
+    store_result_count--;
     return Qnil;
 }
 
@@ -808,8 +880,6 @@ static VALUE error_errno(VALUE obj)
 
 void Init_mysql(void)
 {
-    extern VALUE rb_eStandardError;
-
     cMysql = rb_define_class("Mysql", rb_cObject);
     cMysqlRes = rb_define_class("MysqlRes", rb_cObject);
     cMysqlField = rb_define_class("MysqlField", rb_cObject);
@@ -824,6 +894,9 @@ void Init_mysql(void)
     rb_define_singleton_method(cMysql, "quote", escape_string, 1);
     rb_define_singleton_method(cMysql, "client_info", client_info, 0);
     rb_define_singleton_method(cMysql, "get_client_info", client_info, 0);
+#if MYSQL_VERSION_ID >= 32332
+    rb_define_singleton_method(cMysql, "debug", my_debug, 1);
+#endif
 
     /* Mysql object method */
 #if MYSQL_VERSION_ID >= 32200
@@ -832,14 +905,30 @@ void Init_mysql(void)
     rb_define_method(cMysql, "options", options, -1);
 #endif
     rb_define_method(cMysql, "initialize", initialize, -1);
+#if MYSQL_VERSION_ID >= 32332
+    rb_define_method(cMysql, "escape_string", real_escape_string, 1);
+    rb_define_method(cMysql, "quote", real_escape_string, 1);
+#else
     rb_define_method(cMysql, "escape_string", escape_string, 1);
     rb_define_method(cMysql, "quote", escape_string, 1);
+#endif
     rb_define_method(cMysql, "client_info", client_info, 0);
     rb_define_method(cMysql, "get_client_info", client_info, 0);
     rb_define_method(cMysql, "affected_rows", affected_rows, 0);
+#if MYSQL_VERSION_ID >= 32303
+    rb_define_method(cMysql, "change_user", change_user, -1);
+#endif
+#if MYSQL_VERSION_ID >= 32321
+    rb_define_method(cMysql, "character_set_name", character_set_name, 0);
+#endif
     rb_define_method(cMysql, "close", my_close, 0);
+#if MYSQL_VERSION_ID < 40000
     rb_define_method(cMysql, "create_db", create_db, 1);
     rb_define_method(cMysql, "drop_db", drop_db, 1);
+#endif
+#if MYSQL_VERSION_ID >= 32332
+    rb_define_method(cMysql, "dump_debug_info", dump_debug_info, 0);
+#endif
     rb_define_method(cMysql, "errno", my_errno, 0);
     rb_define_method(cMysql, "error", my_error, 0);
     rb_define_method(cMysql, "field_count", field_count, 0);
@@ -889,9 +978,24 @@ void Init_mysql(void)
 #ifdef REFRESH_STATUS
     rb_define_const(cMysql, "REFRESH_STATUS", INT2NUM(REFRESH_STATUS));
 #endif
+#ifdef REFRESH_THREADS
+    rb_define_const(cMysql, "REFRESH_THREADS", INT2NUM(REFRESH_THREADS));
+#endif
+#ifdef REFRESH_SLAVE
+    rb_define_const(cMysql, "REFRESH_SLAVE", INT2NUM(REFRESH_SLAVE));
+#endif
+#ifdef REFRESH_MASTER
+    rb_define_const(cMysql, "REFRESH_MASTER", INT2NUM(REFRESH_MASTER));
+#endif
+#ifdef CLIENT_LONG_PASSWORD
+#endif
 #ifdef CLIENT_FOUND_ROWS
     rb_define_const(cMysql, "CLIENT_FOUND_ROWS", INT2NUM(CLIENT_FOUND_ROWS));
 #endif
+#ifdef CLIENT_LONG_FLAG
+#endif
+#ifdef CLIENT_CONNECT_WITH_DB
+#endif
 #ifdef CLIENT_NO_SCHEMA
     rb_define_const(cMysql, "CLIENT_NO_SCHEMA", INT2NUM(CLIENT_NO_SCHEMA));
 #endif
@@ -901,6 +1005,27 @@ void Init_mysql(void)
 #ifdef CLIENT_ODBC
     rb_define_const(cMysql, "CLIENT_ODBC", INT2NUM(CLIENT_ODBC));
 #endif
+#ifdef CLIENT_LOCAL_FILES
+    rb_define_const(cMysql, "CLIENT_LOCAL_FILES", INT2NUM(CLIENT_LOCAL_FILES));
+#endif
+#ifdef CLIENT_IGNORE_SPACE
+    rb_define_const(cMysql, "CLIENT_IGNORE_SPACE", INT2NUM(CLIENT_IGNORE_SPACE));
+#endif
+#ifdef CLIENT_CHANGE_USER
+    rb_define_const(cMysql, "CLIENT_CHANGE_USER", INT2NUM(CLIENT_CHANGE_USER));
+#endif
+#ifdef CLIENT_INTERACTIVE
+    rb_define_const(cMysql, "CLIENT_INTERACTIVE", INT2NUM(CLIENT_INTERACTIVE));
+#endif
+#ifdef CLIENT_SSL
+    rb_define_const(cMysql, "CLIENT_SSL", INT2NUM(CLIENT_SSL));
+#endif
+#ifdef CLIENT_IGNORE_SIGPIPE
+    rb_define_const(cMysql, "CLIENT_IGNORE_SIGPIPE", INT2NUM(CLIENT_IGNORE_SIGPIPE));
+#endif
+#ifdef CLIENT_TRANSACTIONS
+    rb_define_const(cMysql, "CLIENT_TRANSACTIONS", INT2NUM(CLIENT_TRANSACTIONS));
+#endif
 
     /* MysqlRes object method */
     rb_define_method(cMysqlRes, "data_seek", data_seek, 1);
@@ -975,6 +1100,12 @@ void Init_mysql(void)
 #ifdef TIMESTAMP_FLAG
     rb_define_const(cMysqlField, "TIMESTAMP_FLAG", INT2NUM(TIMESTAMP_FLAG));
 #endif
+#ifdef SET_FLAG
+    rb_define_const(cMysqlField, "SET_FLAG", INT2NUM(SET_FLAG));
+#endif
+#ifdef NUM_FLAG
+    rb_define_const(cMysqlField, "NUM_FLAG", INT2NUM(NUM_FLAG));
+#endif
 
     /* MysqlError object method */
     rb_define_method(eMysql, "error", error_error, 0);
@@ -1006,4 +1137,37 @@ void Init_mysql(void)
 #ifdef CR_COMMANDS_OUT_OF_SYNC
     rb_define_const(eMysql, "CR_COMMANDS_OUT_OF_SYNC", INT2NUM(CR_COMMANDS_OUT_OF_SYNC));
 #endif
+#ifdef CR_NAMEDPIPE_CONNECTION
+    rb_define_const(eMysql, "CR_NAMEDPIPE_CONNECTION", INT2NUM(CR_NAMEDPIPE_CONNECTION));
+#endif
+#ifdef CR_NAMEDPIPEWAIT_ERROR
+    rb_define_const(eMysql, "CR_NAMEDPIPEWAIT_ERROR", INT2NUM(CR_NAMEDPIPEWAIT_ERROR));
+#endif
+#ifdef CR_NAMEDPIPEOPEN_ERROR
+    rb_define_const(eMysql, "CR_NAMEDPIPEOPEN_ERROR", INT2NUM(CR_NAMEDPIPEOPEN_ERROR));
+#endif
+#ifdef CR_NAMEDPIPESETSTATE_ERROR
+    rb_define_const(eMysql, "CR_NAMEDPIPESETSTATE_ERROR", INT2NUM(CR_NAMEDPIPESETSTATE_ERROR));
+#endif
+#ifdef CR_CANT_READ_CHARSET
+    rb_define_const(eMysql, "CR_CANT_READ_CHARSET", INT2NUM(CR_CANT_READ_CHARSET));
+#endif
+#ifdef CR_NET_PACKET_TOO_LARGE
+    rb_define_const(eMysql, "CR_NET_PACKET_TOO_LARGE", INT2NUM(CR_NET_PACKET_TOO_LARGE));
+#endif
+#ifdef CR_EMBEDDED_CONNECTION
+    rb_define_const(eMysql, "CR_EMBEDDED_CONNECTION", INT2NUM(CR_EMBEDDED_CONNECTION));
+#endif
+#ifdef CR_PROBE_SLAVE_STATUS
+    rb_define_const(eMysql, "CR_PROBE_SLAVE_STATUS", INT2NUM(CR_PROBE_SLAVE_STATUS));
+#endif
+#ifdef CR_PROBE_SLAVE_HOSTS
+    rb_define_const(eMysql, "CR_PROBE_SLAVE_HOSTS", INT2NUM(CR_PROBE_SLAVE_HOSTS));
+#endif
+#ifdef CR_PROBE_SLAVE_CONNECT
+    rb_define_const(eMysql, "CR_PROBE_SLAVE_CONNECT", INT2NUM(CR_PROBE_SLAVE_CONNECT));
+#endif
+#ifdef CR_PROBE_MASTER_CONNECT
+    rb_define_const(eMysql, "CR_PROBE_MASTER_CONNECT", INT2NUM(CR_PROBE_MASTER_CONNECT));
+#endif
 }
